"""
Some plotting functions
"""
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from gym_optimal_intrusion_response.logic.defender_dynamics.defender_dynamics import DefenderDynamics
import gym_optimal_intrusion_response.constants.constants as constants

def plot_alers_logins_distributions() -> None:
    """
    Plots alerts and login distributions

    :return: None
    """
    f1_a = DefenderDynamics.f1_a()
    f1_b = DefenderDynamics.f1_b()
    f2_a = DefenderDynamics.f2_a()
    f2_b = DefenderDynamics.f2_b()

    fontsize = 6.5
    labelsize=10

    plt.rc('text', usetex=True)
    plt.rc('text.latex', preamble=r'\usepackage{amsfonts,amsmath}')
    plt.rcParams['font.family'] = ['serif']
    plt.rcParams['axes.titlepad'] = 0.02
    # plt.rcParams['xtick.major.pad'] = 0.5
    plt.rcParams['ytick.major.pad'] = 0.05
    plt.rcParams['axes.labelpad'] = 0.8
    plt.rcParams['axes.linewidth'] = 0.2
    plt.rcParams.update({'font.size': fontsize})


    fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(6, 1.85))
    x1 = np.arange(0, constants.DP.MAX_ALERTS, 1)
    x1 = np.array(list(filter(lambda k: f1_a.pmf(k) > 0.01, x1.tolist())))
    colors = plt.cm.viridis(np.linspace(0.3, 1, 2))[-2:]
    ax[0].plot(x1, f1_a.pmf(x1), 'ro', ms=5.5, mec=colors[0], color=colors[0])
    ax[0].vlines(x1, 0, f1_a.pmf(x1), label=r"$f_1^A$",
            ls='-', color=colors[0], lw=2)
    x2 = np.arange(0, constants.DP.MAX_ALERTS, 1)
    x2 = np.array(list(filter(lambda k: f1_b.pmf(k) > 0.01, x2.tolist())))
    ax[0].plot(x2, f1_b.pmf(x2), 'ro', ms=5.5, mec="darkred", color="darkred")
    ax[0].vlines(x2, 0, f1_b.pmf(x2), label=r"$f_1^B$",
                  ls='-', color="darkred", lw=2)

    ax[0].set_title(r"Alerts PMFs $f_1^{A}(x), f_1^{B}(x)$", fontsize=11)
    ax[0].set_xlabel(r"\# Alerts $x$", fontsize=8)
    ax[0].grid('on')
    xlab = ax[0].xaxis.get_label()
    ylab = ax[0].yaxis.get_label()
    xlab.set_size(11)
    ylab.set_size(11)
    ax[0].spines['right'].set_color((.8, .8, .8))
    ax[0].spines['top'].set_color((.8, .8, .8))
    ax[0].legend(loc='right', ncol=1)

    ax[0].tick_params(axis='both', which='major', labelsize=7, length=2.2, width=0.6)
    ax[0].tick_params(axis='both', which='minor', labelsize=7, length=2.2, width=0.6)

    x3 = np.arange(0, constants.DP.MAX_LOGINS, 1)
    x3 = np.array(list(filter(lambda k: f2_a.pmf(k) > 0.01, x3.tolist())))
    ax[1].plot(x3, f2_a.pmf(x3), 'ro', ms=5.5, mec=colors[0], color=colors[0])
    ax[1].vlines(x3, 0, f2_a.pmf(x3), label=r"$f_2^A$",
                 ls='-', color=colors[0], lw=2)

    x4 = np.arange(0, constants.DP.MAX_LOGINS, 1)
    x4 = np.array(list(filter(lambda k: f2_b.pmf(k) > 0.01, x4.tolist())))
    ax[1].plot(x4, f2_b.pmf(x4), 'ro', ms=5.5, mec="darkred", color="darkred")
    ax[1].vlines(x4, 0, f2_b.pmf(x4), label=r"$f_2^B$",
                 ls='-', color="darkred", lw=2)

    ax[1].set_title(r"Login Attempts PMFs $f_2^{A}(y), f_2^{B}(y)$", fontsize=11)
    ax[1].set_xlabel(r"\# Login attempts $y$", fontsize=8)

    ax[1].tick_params(axis='both', which='major', labelsize=7, length=2.2, width=0.6)
    ax[1].tick_params(axis='both', which='minor', labelsize=7, length=2.2, width=0.6)
    ax[1].grid('on')
    xlab = ax[1].xaxis.get_label()
    ylab = ax[1].yaxis.get_label()
    xlab.set_size(11)
    ylab.set_size(11)
    ax[1].spines['right'].set_color((.8, .8, .8))
    ax[1].spines['top'].set_color((.8, .8, .8))
    ax[1].legend(loc='right', ncol=1)

    fig.tight_layout()
    # plt.show()
    plt.subplots_adjust(wspace=0.10)
    fig.savefig("alerts_logins_distributions" + ".png", format="png", dpi=600)
    fig.savefig("alerts_logins_distributions" + ".pdf", format='pdf', dpi=600, bbox_inches='tight', transparent=True)
    # plt.close(fig)


def ttc_plot_helper(x, y, max_x) -> np.ndarray:
    """
    Utility for plotting the TTC

    :param x: number of alerts
    :param y: number of login attempts
    :param max_x: max number of alerts
    :return: the TTC values
    """
    z = []
    for i in range(len(x)):
        z1 = []
        for j in range(len(x[i])):
            val = DefenderDynamics.ttc(x[i][j], y[i, j], max_x)
            z1.append(val)
        z.append(z1)
    z = np.array(z)
    return z


def hack_prob_helper(ttc_val, t) -> np.ndarray:
    """
    Utility function for plotting the hack probability

    :param ttc_val: the TTC value
    :param t: the time-step
    :return: the hack probabilities
    """
    z = []
    for i in range(len(ttc_val)):
        z1 = []
        for j in range(len(ttc_val[i])):
            val = DefenderDynamics.hack_prob(ttc_val[i][j], t[i][j])
            z1.append(val)
        z.append(z1)
    z = np.array(z)
    return z


def plot_ttc() -> None:
    """
    Plots the TTC

    :return: None
    """
    v = np.arange(1, constants.DP.MAX_ALERTS, 1)
    s = np.arange(1, constants.DP.MAX_LOGINS, 1)
    x,y = np.meshgrid(v,  s)
    z = ttc_plot_helper(x,y,constants.DP.MAX_ALERTS)

    plt.rc('text', usetex=True)
    plt.rc('text.latex', preamble=r'\usepackage{amsfonts,amsmath}')
    plt.rcParams['font.family'] = ['serif']
    plt.rcParams['axes.titlepad'] = 15
    plt.rcParams['xtick.major.pad'] = 0.05
    plt.rcParams['ytick.major.pad'] = 0.05
    plt.rcParams['axes.labelpad'] = 0.05
    plt.rcParams['axes.linewidth'] = 0.2
    plt.rcParams.update({'font.size': 6.5})
    plt.rc('text', usetex=True)
    plt.rc('text.latex', preamble=r'\usepackage{amsfonts}')
    plt.rcParams['font.family'] = ['serif']
    plt.rcParams['font.serif'] = ['Times New Roman']
    plt.rcParams.update({'font.size': 10})
    fig, ax = plt.subplots(nrows=1, ncols=1, subplot_kw={'projection': '3d'}, figsize=(3.85, 2.7))
    ax.plot_surface(x, y, z, cmap='viridis_r', linewidth=0.1,
                    alpha=0.8, edgecolor='k', rstride=2, cstride=2)

    ax.set_title(r"$TTC(x,y)$", fontsize=11)
    ax.set_xlabel(r"IDS alerts $x$")
    ax.set_ylabel(r"Login attempts $y$")
    xlab = ax.xaxis.get_label()
    ylab = ax.yaxis.get_label()
    xlab.set_size(8)
    ylab.set_size(8)
    ax.tick_params(axis='both', which='major', labelsize=7, length=3.2, width=0.1)
    ax.tick_params(axis='both', which='minor', labelsize=7, length=3.2, width=0.1)
    plt.ylim(constants.DP.MAX_LOGINS, 0.0)
    fig.tight_layout()
    # plt.show()
    fig.savefig("ttc_alerts_logins" + ".png", format="png", dpi=600)
    fig.savefig("ttc_alerts_logins" + ".pdf", format='pdf', dpi=600, bbox_inches='tight', transparent=True)
    # plt.close(fig)


def plot_hp() -> None:
    """
    Plots the hack probability

    :return: None
    """
    t = np.arange(1, constants.DP.MAX_TIMESTEPS,1)
    z = np.arange(1, constants.DP.MAX_TTC, 1)
    x2,y2 = np.meshgrid(z, t)
    z2 = hack_prob_helper(x2, y2)


    plt.rc('text', usetex=True)
    plt.rc('text.latex', preamble=r'\usepackage{amsfonts,amsmath}')
    plt.rcParams['font.family'] = ['serif']
    plt.rcParams['axes.titlepad'] = 15
    plt.rcParams['xtick.major.pad'] = 0.05
    plt.rcParams['ytick.major.pad'] = 0.05
    plt.rcParams['axes.labelpad'] = 0.2
    plt.rcParams['axes.linewidth'] = 0.2
    plt.rcParams.update({'font.size': 6.5})
    plt.rc('text', usetex=True)
    plt.rc('text.latex', preamble=r'\usepackage{amsfonts}')
    plt.rcParams['font.family'] = ['serif']
    plt.rcParams['font.serif'] = ['Times New Roman']
    plt.rcParams.update({'font.size': 10})
    fig, ax = plt.subplots(nrows=1, ncols=1, subplot_kw={'projection': '3d'},  figsize=(3.85, 2.7))
    ax.plot_surface(x2, y2, z2, cmap='viridis_r', linewidth=0.1,
                    alpha=0.8, edgecolor='k', rstride=2, cstride=2)

    ax.set_title(r"Probability of intrusion $\phi(x_t, y_t, t)$", fontsize=11)
    ax.set_xlabel(r"$TTC(x_t, y_t)$")
    ax.set_ylabel(r"Time $t$")
    xlab = ax.xaxis.get_label()
    ylab = ax.yaxis.get_label()
    xlab.set_size(8)
    ylab.set_size(8)
    ax.tick_params(axis='both', which='major', labelsize=7, length=2.2, width=0.6)
    ax.tick_params(axis='both', which='minor', labelsize=7, length=2.2, width=0.6)
    fig.tight_layout()
    # plt.show()
    fig.savefig("intrusion_prob" + ".png", format="png", dpi=600)
    fig.savefig("intrusion_prob" + ".pdf", format='pdf', dpi=600, bbox_inches='tight', transparent=True)
    # plt.close(fig)


def plot_hp_2() -> None:
    """
    Plots the hack probability

    :return: None
    """
    x = np.arange(1, constants.DP.MAX_TTC, 1)
    y = list(map(lambda g: DefenderDynamics.hack_prob(g, 10), x))


    plt.rc('text', usetex=True)
    plt.rc('text.latex', preamble=r'\usepackage{amsfonts,amsmath}')
    plt.rcParams['font.family'] = ['serif']
    plt.rcParams['axes.titlepad'] = 0.02
    # plt.rcParams['xtick.major.pad'] = 0.5
    plt.rcParams['ytick.major.pad'] = 0.05
    plt.rcParams['axes.labelpad'] = 0.8
    plt.rcParams['axes.linewidth'] = 0.1
    plt.rcParams.update({'font.size': 10})
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(3.5, 3.2))

    # Plot Avg Eval rewards Gensim
    colors = plt.cm.viridis(np.linspace(0.3, 1, 2))[-2:]
    ax.plot(x,
            y, label=r"$\pi_{\theta}$ simulation",
            ls='-', color=colors[0])
    ax.fill_between(x, y, np.zeros(len(y)),
                    alpha=0.35, color=colors[0])
    ax.set_title(r"$\mathbb{P}[\text{intrusion} | c]$", fontsize=12.5)
    ax.set_xlabel(r"TTC $c$", fontsize=11.5)
    ax.set_xlim((0, 75))

    # set the grid on
    ax.grid('on')

    # tweak the axis labels
    xlab = ax.xaxis.get_label()
    ylab = ax.yaxis.get_label()

    xlab.set_size(11.5)
    ylab.set_size(11.5)

    # change the color of the top and right spines to opaque gray
    ax.spines['right'].set_color((.8, .8, .8))
    ax.spines['top'].set_color((.8, .8, .8))

    ttl = ax.title
    ttl.set_position([.5, 1.05])

    fig.tight_layout()
    plt.show()


if __name__ == '__main__':
    plot_ttc()
    plot_hp()
    plot_alers_logins_distributions()
    # plot_intrusion_dist()